#!/bin/sh

# Copyright (C) 2006 Kel Modderman <kelrin@tpg.com.au>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# On Debian GNU/Linux systems, the text of the GPL license can be
# found in /usr/share/common-licenses/GPL.

# Quit if we're called for the loopback
if [ "$IFACE" = lo ]; then
	exit 0
fi

# Quit if wpa_action is active
if [ -f "/var/run/wpa_action.$IFACE.pid" ]; then
	exit 0
fi

WPA_SUP_BIN="/sbin/wpa_supplicant"
WPA_SUP_PNAME="wpa_supplicant"

WPA_CLI_BIN="/sbin/wpa_cli"
WPA_CLI_PNAME="wpa_cli"

# quit if executables are not installed
if [ ! -x "$WPA_SUP_BIN" -o ! -x "$WPA_CLI_BIN" ]; then
	exit 0
fi

# wpa_supplicant pidfile
WPA_SUP_PIDFILE="/var/run/wpa_supplicant.$IFACE.pid"

# wpa_cli daemon pidfile name depends on how it is invoked
if [ -n "$IF_WPA_ROAM" ]; then
	if [ "$METHOD" != "manual" ]; then
		echo "$WPA_SUP_PNAME, wpa-roam can only be used with the \"manual\" inet METHOD" >/dev/stderr
		exit 1
	fi
	# this pidfile will make ifupdown.sh no-op on consequent if{up,down}
	# operations performed for IFACE
	WPA_CLI_PIDFILE="/var/run/wpa_action.$IFACE.pid"
	# shunt roam conf to IF_WPA_CONF
	IF_WPA_CONF="$IF_WPA_ROAM"
	# action script is wpa_action
	WPA_ACTION_SCRIPT="/sbin/wpa_action"
else
	# default wpa_cli daemon pidfile
	WPA_CLI_PIDFILE="/var/run/wpa_cli.$IFACE.pid"
	# user supplied wpa_cli event action script
	if [ -n "$IF_WPA_ACTION" ]; then
		WPA_ACTION_SCRIPT="$IF_WPA_ACTION"
	fi
fi

# path to common ctrl_interface socket for active interfaces
# this is the default path provided by upstream
WPA_CTRL_DIR="/var/run/wpa_supplicant"

# method of determining the outcome of an user supplied action script
# see /usr/share/doc/wpasupplicant/examples/wpacli-action-skeleton
WPA_CLI_ACTFILE="/var/run/wpa_cli.$IFACE.action"

# master function; determines if ifupdown.sh should do something or not
if [ -s "$IF_WPA_CONF" ]; then
	# if a conf is provided and it has some content, use it
	WPA_SUP_CONF="-c $IF_WPA_CONF"
	# get ctrl_interface socket path defined in conf
	WPA_SUP_CONF_CTRL_DIR=$(sed -n -e 's/[ \t]*#.*//g;s/[ \t]*$//g;s/^ctrl_interface=//p' "$IF_WPA_CONF")
	# use ctrl_interface socket path when defined in user supplied conf, overriding the pre-set default
	if [ -n "$WPA_SUP_CONF_CTRL_DIR" ]; then
		WPA_CTRL_DIR="$WPA_SUP_CONF_CTRL_DIR"
	fi
elif set | grep -q "^IF_WPA"; then
	# any wpa-* line causes wpa_supplicant to launch with an unconfigured ctrl_interface socket
	# the socket is later configured by a series of set_network cmd's in conf_wpa_supplicant()
	WPA_SUP_CONF="-C $WPA_CTRL_DIR"
else
	exit 0
fi

if [ -n "$IF_WPA_VERBOSITY" -o "$VERBOSITY" = "1" ]; then
	# show wpa_cli messages when VERBOSITY has been set
	TO_NULL="/dev/stdout"
	# start-stop-daemon verbosity
	DAEMON_VEROSITY="--verbose"
else
	# hide infomational messages when VERBOSITY is not set
	TO_NULL="/dev/null"
	DAEMON_VERBOSITY="--quiet"
fi

# wpa_cli wrapper function
# avoid language related interferences
# supply path to common ctrl_interface socket(s)
# supply IFACE as argument
wpa_cli () {
	LC_ALL=C $WPA_CLI_BIN -p $WPA_CTRL_DIR -i $IFACE "$@"
}

# wpa_cli_do <value> <type> <variable> [set_network variable] <desc>
# process environment variables resulting from wpa-* lines in the interfaces file
# $1 = envorinment variable
# $2 = data type of variable {raw|ascii}
# $3 = wpa_cli variable, if $3 is set_network, shift and take set_network subvariable
# $4 = wpa-* string as it would appear in interfaces file, enhances information messages
wpa_cli_do () {
	
	if [ -z "$1" ]; then
		# environment variable not defined, no-op
		return
	fi
	
	local WPACLISET_VALUE WPACLISET_VARIABLE WPACLISET_DESC
	
	# provide a method to disable the forced choice of ascii data type
	# using this option requires that strings are quoted as required in interfaces file
	if [ -n "$IF_WPA_INPUT_TYPE" ]; then
		WPACLISET_VALUE="$1"
	else
		# wpa_cli value and type
		case "$2" in
			ascii)
				# Double quote
				WPACLISET_VALUE="\"$1\""
				;;
			raw|*)
				# Provide raw value
				WPACLISET_VALUE="$1"
				;;
		esac
	fi
	
	# wpa_cli variable
	case "$3" in
		set_network)
			if [ -z "$WPA_ID" ]; then
				# WPA_ID is required for set_network commands
				return
			fi
			# Shift to set_network variable
			shift
			WPACLISET_VARIABLE="set_network $WPA_ID $3"
			;;
		*)
			WPACLISET_VARIABLE="$3"
			;;
	esac
	
	# display description of wpa_cli command, censor sensitive knowledge
	case "$4" in
		*-psk|*-passphrase|*-passwd*|*-wep-key*)
			# For sensitive knowledge
			WPACLISET_DESC="$4 *****"
			;;
		*)
			# Echo actual value used
			WPACLISET_DESC="$4 $WPACLISET_VALUE"
			;;
	esac

	# echo debug/verbosity messages
	echo -n "$WPA_SUP_PNAME, $WPACLISET_DESC : " >$TO_NULL
	
	# NB: $WPACLISET_VARIABLE must be unprotected and $WPACLISET_VALUE must be protected
	wpa_cli $WPACLISET_VARIABLE "$WPACLISET_VALUE" >$TO_NULL
}

init_wpa_supplicant () {
	if [ ! -S "$WPA_CTRL_DIR/$IFACE" ]; then
		
		# action script sanity checking
		if [ -n "$WPA_ACTION_SCRIPT" ]; then
			if [ -x "$WPA_ACTION_SCRIPT" ]; then
				# wait for wpa_cli to attach
				WPA_SUP_OPTIONS="-W -B -P $WPA_SUP_PIDFILE -i $IFACE"	
			else
				echo "$WPA_SUP_PNAME, action script \"$WPA_ACTION_SCRIPT\" not executable, aborting" >/dev/stderr
				exit 1
			fi
		else
			# default options, daemonise, create pidfile, supply IFACE
			WPA_SUP_OPTIONS="-B -P $WPA_SUP_PIDFILE -i $IFACE"
		fi

		# driver type of interface, defaults to wext when undefined
		if [ -n "$IF_WPA_DRIVER" ]; then
			WPA_DRIVER="$IF_WPA_DRIVER"
			echo "$WPA_SUP_PNAME, wpa-driver $WPA_DRIVER" >$TO_NULL
		else
			# wext is default/fallback
			WPA_DRIVER="wext"
			echo "$WPA_SUP_PNAME, using default driver type: wpa-driver $WPA_DRIVER" >$TO_NULL
		fi

		# if we have passed the criteria, start wpa_supplicant
		if [ -n "$WPA_SUP_CONF" ]; then
			
			echo "$WPA_SUP_PNAME, $WPA_SUP_BIN $WPA_SUP_OPTIONS $WPA_SUP_CONF -D $WPA_DRIVER" >$TO_NULL
			
			start-stop-daemon --start --oknodo $DAEMON_VEROSITY \
				--name $WPA_SUP_PNAME --startas $WPA_SUP_BIN --pidfile $WPA_SUP_PIDFILE \
				-- $WPA_SUP_OPTIONS $WPA_SUP_CONF -D $WPA_DRIVER
		fi
	else
		echo "$WPA_SUP_PNAME, unknown process created $WPA_CTRL_DIR/$IFACE, exiting" >/dev/stderr
		exit 1
	fi
}

conf_wpa_supplicant () {
	
	# if the interface socket exists, then wpa_supplicant was invoked successfully
	if [ -S "$WPA_CTRL_DIR/$IFACE" ]; then
		echo "$WPA_SUP_PNAME, ctrl_interface socket located at $WPA_CTRL_DIR/$IFACE" >$TO_NULL
	
		# global settings
		wpa_cli_do "$IF_WPA_AP_SCAN" raw \
			ap_scan wpa-ap-scan
		
		wpa_cli_do "$IF_WPA_PREAUTHENTICATE" raw \
			preauthenticate wpa-preauthenticate
		
		# per ssid specific settings
		if [ -n "$IF_WPA_SSID" ]; then
			
			local WPA_ID
			
			# initialize a new network block, record the unique WPA_ID
			WPA_ID=$(wpa_cli add_network)
			
			echo "$WPA_SUP_PNAME, configuring new network block : $WPA_ID" >$TO_NULL
			
			# set an ssid
			wpa_cli_do "$IF_WPA_SSID" ascii \
				set_network ssid wpa-ssid
			
			# configure the new network block
			wpa_cli_do "$IF_WPA_PRIORITY" raw \
				set_network priority wpa-priority
			
			wpa_cli_do "$IF_WPA_BSSID" raw \
				set_network bssid wpa-bssid
			
			if [ -s "$IF_WPA_PSK_FILE" ]; then
				IF_WPA_PSK=$(cat "$IF_WPA_PSK_FILE")
			fi
			wpa_cli_do "$IF_WPA_PSK" raw \
				set_network psk wpa-psk
			
			if [ -s "$IF_WPA_PASSPHRASE_FILE" ]; then
				IF_WPA_PASSPHRASE=$(cat "$IF_WPA_PASSPHRASE_FILE")
			fi
			wpa_cli_do "$IF_WPA_PASSPHRASE" ascii \
				set_network psk wpa-passphrase
			
			wpa_cli_do "$IF_WPA_PAIRWISE" raw \
				set_network pairwise wpa-pairwise
			
			wpa_cli_do "$IF_WPA_GROUP" raw \
				set_network group wpa-group
			
			wpa_cli_do "$IF_WPA_KEY_MGMT" raw \
				set_network key_mgmt wpa-key-mgmt
			
			wpa_cli_do "$IF_WPA_PROTO" raw \
				set_network proto wpa-proto
			
			wpa_cli_do "$IF_WPA_AUTH_ALG" raw \
				set_network auth_alg wpa-auth-alg
			
			wpa_cli_do "$IF_WPA_SCAN_SSID" raw \
				set_network scan_ssid wpa-scan-ssid
			
			wpa_cli_do "$IF_WPA_IDENTITY" ascii \
				set_network identity wpa-identity
			
			wpa_cli_do "$IF_WPA_ANONYMOUS_IDENTITY" ascii \
				set_network anonymous_identity wpa-anonymous-identity
			
			wpa_cli_do "$IF_WPA_EAP" raw \
				set_network eap wpa-eap
			
			wpa_cli_do "$IF_WPA_EAPPSK" raw \
				set_network eappsk wpa-eappsk

			wpa_cli_do "$IF_WPA_NAI" ascii \
				set_network nai wpa-nai

			wpa_cli_do "$IF_WPA_PASSWORD" ascii \
				set_network password wpa-password

			wpa_cli_do "$IF_WPA_CA_CERT" ascii \
				set_network ca_cert wpa-ca-cert

			wpa_cli_do "$IF_WPA_CA_PATH" ascii \
				set_network ca_path wpa-ca-path

			wpa_cli_do "$IF_WPA_CLIENT_CERT" ascii \
				set_network client_cert wpa-client-cert

			wpa_cli_do "$IF_WPA_PRIVATE_KEY" ascii \
				set_network private_key wpa-private-key

			wpa_cli_do "$IF_WPA_PRIVATE_KEY_PASSWD" ascii \
				set_network private_key_passwd wpa-private-key-passwd
			
			wpa_cli_do "$IF_WPA_DH_FILE" ascii \
				set_network dh_file wpa-dh-file

			wpa_cli_do "$IF_WPA_SUBJECT_MATCH" ascii \
				set_network subject_match wpa-subject-match

			wpa_cli_do "$IF_WPA_ALTSUBJECT_MATCH" ascii \
				set_network altsubject_match wpa-altsubject-match

			wpa_cli_do "$IF_WPA_CA_CERT2" ascii \
				set_network ca_cert2 wpa-ca-cert2

			wpa_cli_do "$IF_WPA_CA_PATH2" ascii \
				set_network ca_path2 wpa-ca-path2

			wpa_cli_do "$IF_WPA_CLIENT_CERT2" ascii \
				set_network client_cert2 wpa-client-cert2

			wpa_cli_do "$IF_WPA_PRIVATE_KEY2" ascii \
				set_network private_key2 wpa-private-key2

			wpa_cli_do "$IF_WPA_PRIVATE_KEY_PASSWD2" ascii \
				set_network private_key_passwd2 wpa-private-key-passwd2
			
			wpa_cli_do "$IF_WPA_DH_FILE2" ascii \
				set_network dh_file2 wpa-dh-file2

			wpa_cli_do "$IF_WPA_SUBJECT_MATCH2" ascii \
				set_network subject_match2 wpa-subject-match2

			wpa_cli_do "$IF_WPA_ALTSUBJECT_MATCH2" ascii \
				set_network altsubject_match2 wpa-altsubject-match2
			
			wpa_cli_do "$IF_WPA_EAP_METHODS" raw \
				set_network eap_methods wpa-eap-methods

			wpa_cli_do "$IF_WPA_PHASE1" ascii \
				set_network phase1 wpa-phase1

			wpa_cli_do "$IF_WPA_PHASE2" ascii \
				set_network phase2 wpa-phase2

			wpa_cli_do "$IF_WPA_PCSC" raw \
				set_network pcsc wpa-pcsc

			wpa_cli_do "$IF_WPA_PIN" ascii \
				set_network pin wpa-pin

			wpa_cli_do "$IF_WPA_ENGINE" raw \
				set_network engine wpa-engine

			wpa_cli_do "$IF_WPA_ENGINE_ID" ascii \
				set_network engine_id wpa-engine-id

			wpa_cli_do "$IF_WPA_KEY_ID" ascii \
				set_network key_id wpa-key-id

			wpa_cli_do "$IF_WPA_EAPOL_FLAGS" raw \
				set_network eapol_flags wpa-eapol-flags
			
			wpa_cli_do "$IF_WPA_WEP_KEY0" raw \
				set_network wep_key0 wpa-wep-key0
			
			wpa_cli_do "$IF_WPA_WEP_KEY1" raw \
				set_network wep_key1 wpa-wep-key1
			
			wpa_cli_do "$IF_WPA_WEP_KEY2" raw \
				set_network wep_key2 wpa-wep-key2
			
			wpa_cli_do "$IF_WPA_WEP_KEY3" raw \
				set_network wep_key3 wpa-wep-key3
			
			wpa_cli_do "$IF_WPA_WEP_TX_KEYIDX" raw \
				set_network wep_tx_keyidx wpa-wep-tx-keyidx
			
			wpa_cli_do "$IF_WPA_PROACTIVE_KEY_CACHING" raw \
				set_network proactive_key_caching wpa-proactive-key-caching
			
			wpa_cli_do "$IF_WPA_PAC_FILE" ascii \
				set_network pac_file wpa-pac-file
			
			wpa_cli_do "$IF_WPA_MODE" raw \
				set_network mode wpa-mode
			
			wpa_cli_do "$IF_WPA_STAKEY" raw \
				set_network stakey wpa-stakey
			
			wpa_cli_do "$IF_WPA_PEERKEY" raw \
				set_network peerkey wpa-peerkey
			
			wpa_cli_do "$IF_FRAGMENT_SIZE" raw \
				set_network fragment_size wpa-fragment-size

			wpa_cli_do "$IF_WPA_ID_STR" ascii \
				set_network id_str wpa-id-str
			
			# finality: activate the new network block
			wpa_cli_do "$WPA_ID" raw \
				enable_network "enabling network block"
		fi
	else
		# wpa_supplicant socket does not exist; wpa_supplicant failed to start
		echo "$WPA_SUP_PNAME, failed to locate ctrl_interface socket, exiting" >/dev/stderr
		exit 1
	fi
}

kill_wpa_supplicant () {
	if [ -f "$WPA_SUP_PIDFILE" ]; then
		
		echo "$WPA_SUP_PNAME, terminating $WPA_SUP_PNAME daemon via pidfile $WPA_SUP_PIDFILE" >$TO_NULL

		start-stop-daemon --stop --oknodo $DAEMON_VEROSITY \
			--name $WPA_SUP_PNAME --pidfile	$WPA_SUP_PIDFILE
	
		if [ -f "$WPA_SUP_PIDFILE" ]; then
			rm -f "$WPA_SUP_PIDFILE"
		fi
	elif [ -S "$WPA_CTRL_DIR/$IFACE" ]; then

		# terminate existing ctrl_interface socket
		echo -n "$WPA_SUP_PNAME, terminating via ctrl_iface socket $WPA_CTRL_DIR/$IFACE : " >$TO_NULL

		wpa_cli terminate >$TO_NULL
		
		# remove stale socket
		if [ -S "$WPA_CTRL_DIR/$IFACE" ]; then
			rm -f "$WPA_CTRL_DIR/$IFACE"
		fi
	fi
}

init_wpa_cli () {
	if [ -n "$WPA_ACTION_SCRIPT" ]; then

		# default wpa_cli daemon options
		# daemonise, create pidfile, supply IFACE
		WPA_CLI_OPTIONS="-B -P $WPA_CLI_PIDFILE -i $IFACE"

		# ensure the actfile does not exist
		if [ -f "$WPA_CLI_ACTFILE" ]; then
			rm -f "$WPA_CLI_ACTFILE"
		fi
			
		echo "$WPA_SUP_PNAME, $WPA_CLI_BIN $WPA_CLI_OPTIONS -p $WPA_CTRL_DIR -a $WPA_ACTION_SCRIPT" >$TO_NULL
			
		start-stop-daemon --start --oknodo $DAEMON_VEROSITY \
			--name $WPA_CLI_PNAME --startas $WPA_CLI_BIN --pid $WPA_CLI_PIDFILE \
			-- $WPA_CLI_OPTIONS -p $WPA_CTRL_DIR -a $WPA_ACTION_SCRIPT
	fi
}

conf_wpa_cli () {
	if [ -n "$IF_WPA_ACTION" -a "$METHOD" = "manual" ]; then
		
		# IF_WPA_ACTION timeout loop
		WPA_ACTION_TIME="0"
		if [ -z "$IF_WPA_ACTION_TIMEOUT" ]; then
			# default: allow one minute for action script to succeed
			IF_WPA_ACTION_TIMEOUT="60"
		fi
		
		while true; do
			if [ "$IF_WPA_ACTION_TIMEOUT" -lt "0" ]; then
				# dont wait
				echo "$WPA_SUP_PNAME, backgrounding $IF_WPA_ACTION action script" >/dev/stdout
				break
			elif [ -e "$WPA_CLI_ACTFILE" ]; then
				# success
				echo "$WPA_SUP_PNAME, action script $IF_WPA_ACTION succeeded" >/dev/stdout
				break
			elif [ "$IF_WPA_ACTION_TIMEOUT" -gt "0" -a "$WPA_ACTION_TIME" -ge "$IF_WPA_ACTION_TIMEOUT" ]; then
				# NB: it is safe to check for wpa_cli's pidfile only after sufficient time has passed
				if [ ! -e "$WPA_CLI_PIDFILE" ]; then
					# failure
					echo "$WPA_SUP_PNAME, action script failed" >/dev/stderr
					exit 1
				else 
					# timeout
					echo "$WPA_SUP_PNAME, action script timed out, backgrounding" >/dev/stdout
					break
				fi
			fi
			
			WPA_ACTION_TIME=$(($WPA_ACTION_TIME+1))
			sleep 1
		done
	fi
}

kill_wpa_cli () {
	if [ -f "$WPA_CLI_PIDFILE" ]; then
		
		echo -n "$WPA_SUP_PNAME, sending disconnect signal to $WPA_CLI_PNAME daemon : " >$TO_NULL
		
		# send DISCONNECTED signal first
		# allow the action script a chance to undo network settings
		wpa_cli disconnect >$TO_NULL

		echo "$WPA_SUP_PNAME, terminating $WPA_CLI_PNAME daemon via pidfile $WPA_CLI_PIDFILE" >$TO_NULL
		
		start-stop-daemon --stop --oknodo $DAEMON_VEROSITY \
			--name $WPA_CLI_PNAME --pidfile	$WPA_CLI_PIDFILE
		
		if [ -f "$WPA_CLI_PIDFILE" ]; then
			rm -f "$WPA_CLI_PIDFILE"
		fi

		if [ -f "$WPA_CLI_ACTFILE" ]; then
			rm -f "$WPA_CLI_ACTFILE"
		fi
	fi
}

case "$MODE" in 
	start)
		case "$PHASE" in
			pre-up)
				kill_wpa_supplicant
				init_wpa_supplicant
				init_wpa_cli
				conf_wpa_supplicant
				conf_wpa_cli
				;;
			*)
				echo "$WPA_SUP_PNAME, unknown phase: \"$PHASE\"" >/dev/stderr
				exit 1
				;;
		esac
		;;

	stop)
		case "$PHASE" in
			pre-down)
				kill_wpa_cli
				;;
			post-down)
				kill_wpa_supplicant
				;;
			*)
				echo "$WPA_SUP_PNAME, unknown phase: \"$PHASE\"" >/dev/stderr
				exit 1
				;;
		esac
		;;
	
	*)
		echo "$WPA_SUP_PNAME, unknown mode: \"$MODE\"" >/dev/stderr
		exit 1
		;;
esac
